from .envs import *
import pytest
from crypticorn.common import (
    AuthHandler,
    Scope,
    HTTPException,
    ApiError,
    WebSocketException,
)
from fastapi.security import HTTPAuthorizationCredentials


# ASSERT SCOPE
PURCHASEABLE_SCOPES = Scope.purchaseable_scopes()
ADMIN_SCOPES = Scope.admin_scopes()
INTERNAL_SCOPES = Scope.internal_scopes()

# Debug
UPDATE_SCOPES = "you probably need to bring the scopes in both the api client and the auth service in sync"

# Each function is tested without credentials, with invalid credentials, and with valid credentials.
# The test is successful if the correct HTTPException is raised.


# COMBINED AUTH
@pytest.mark.asyncio
async def test_combined_auth_without_credentials(auth_handler: AuthHandler):
    """Without credentials"""
    with pytest.raises(HTTPException) as e:
        await auth_handler.combined_auth(bearer=None, api_key=None)
    assert e.value.status_code == 401
    assert e.value.detail.get("code") == ApiError.NO_CREDENTIALS.identifier


# BEARER
@pytest.mark.asyncio
async def test_combined_auth_with_invalid_bearer_token(auth_handler: AuthHandler):
    """With invalid bearer token"""
    with pytest.raises(HTTPException) as e:
        await auth_handler.combined_auth(
            bearer=HTTPAuthorizationCredentials(scheme="Bearer", credentials="123"),
            api_key=None,
        )
    assert e.value.status_code == 401
    assert e.value.detail.get("code") == ApiError.INVALID_BEARER.identifier


@pytest.mark.asyncio
async def test_combined_auth_with_expired_bearer_token(auth_handler: AuthHandler):
    """With expired bearer token"""
    with pytest.raises(HTTPException) as e:
        await auth_handler.combined_auth(
            bearer=HTTPAuthorizationCredentials(
                scheme="Bearer", credentials=EXPIRED_JWT
            ),
            api_key=None,
        )
    assert e.value.status_code == 401
    assert e.value.detail.get("code") == ApiError.EXPIRED_BEARER.identifier


@pytest.mark.asyncio
async def test_combined_auth_with_valid_bearer_token(auth_handler: AuthHandler):
    """With valid bearer token"""
    res = await auth_handler.combined_auth(
        bearer=HTTPAuthorizationCredentials(scheme="Bearer", credentials=VALID_JWT),
        api_key=None,
    )
    assert all(
        [key not in res.scopes for key in PURCHASEABLE_SCOPES]
    ), "non admin should not have access to purchaseable scopes"
    assert all(
        [key not in res.scopes for key in ADMIN_SCOPES]
    ), "non admin should not have access to any of the admin keys"
    assert all(
        [key not in res.scopes for key in INTERNAL_SCOPES]
    ), "non admin should not have access to any of the internal keys"
    assert not res.admin, "non admin should not be admin"


@pytest.mark.asyncio
async def test_combined_auth_with_valid_prediction_bearer_token(
    auth_handler: AuthHandler,
):
    """With valid bearer token"""
    res = await auth_handler.combined_auth(
        bearer=HTTPAuthorizationCredentials(
            scheme="Bearer", credentials=VALID_PREDICTION_JWT
        ),
        api_key=None,
    )
    assert all(
        [key in res.scopes for key in PURCHASEABLE_SCOPES]
    ), "non admin which purchased predictions should have access to purchaseable scopes"
    assert all(
        [key not in res.scopes for key in ADMIN_SCOPES]
    ), "non admin should not have access to any of the admin keys"
    assert all(
        [key not in res.scopes for key in INTERNAL_SCOPES]
    ), "non admin should not have access to any of the internal keys"
    assert not res.admin, "non admin should not be admin"


@pytest.mark.asyncio
async def test_combined_auth_with_valid_admin_bearer_token(auth_handler: AuthHandler):
    """With valid admin bearer token"""
    res = await auth_handler.combined_auth(
        bearer=HTTPAuthorizationCredentials(
            scheme="Bearer", credentials=VALID_ADMIN_JWT
        ),
        api_key=None,
    )
    assert all(
        [key in res.scopes for key in PURCHASEABLE_SCOPES]
    ), "admin should have access to purchaseable scopes"
    assert all(
        [key in res.scopes for key in ADMIN_SCOPES]
    ), "admin should have access to any of the admin keys"
    assert all(
        [key not in res.scopes for key in INTERNAL_SCOPES]
    ), "admin should not have access to any of the internal keys"
    assert res.admin, "admin should be true"


# API KEY
@pytest.mark.asyncio
async def test_combined_auth_with_invalid_api_key(auth_handler: AuthHandler):
    """With invalid api key"""
    with pytest.raises(HTTPException) as e:
        return await auth_handler.combined_auth(bearer=None, api_key="123")
    assert e.value.status_code == 401
    assert e.value.detail.get("code") == ApiError.INVALID_API_KEY.identifier


@pytest.mark.asyncio
async def test_combined_auth_with_full_scope_valid_api_key(auth_handler: AuthHandler):
    """With full scope valid api key"""
    res = await auth_handler.combined_auth(bearer=None, api_key=FULL_SCOPE_API_KEY)
    assert res.scopes == [
        x for x in Scope if x not in Scope.internal_scopes()
    ], UPDATE_SCOPES


@pytest.mark.asyncio
async def test_combined_auth_with_one_scope_valid_api_key(auth_handler: AuthHandler):
    """With one scope valid api key"""
    res = await auth_handler.combined_auth(bearer=None, api_key=ONE_SCOPE_API_KEY)
    assert ONE_SCOPE_API_KEY_SCOPE in res.scopes, UPDATE_SCOPES
    assert len(res.scopes) == 1, "should only have one scope"


@pytest.mark.asyncio
async def test_combined_auth_with_expired_api_key(auth_handler: AuthHandler):
    """With expired api key"""
    with pytest.raises(HTTPException) as e:
        res = await auth_handler.combined_auth(bearer=None, api_key=EXPIRED_API_KEY)
    assert e.value.status_code == 401
    assert e.value.detail.get("code") == ApiError.EXPIRED_API_KEY.identifier


# WS COMBINED AUTH
@pytest.mark.asyncio
async def test_ws_combined_auth_websocket_exception(auth_handler: AuthHandler):
    """Without credentials"""
    with pytest.raises(WebSocketException) as e:
        return await auth_handler.ws_combined_auth(bearer=None, api_key=None)
    assert e.value.status_code == 1008
    assert e.value.detail.get("code") == ApiError.NO_CREDENTIALS.identifier



from typing import TypeVar
from crypticorn.hive import HiveClient
from crypticorn.klines import KlinesClient
from crypticorn.pay import PayClient
from crypticorn.trade import TradeClient
from crypticorn.metrics import MetricsClient
from crypticorn.auth import AuthClient
from crypticorn.common import BaseUrl, ApiVersion, Service, apikey_header as aph

ConfigT = TypeVar("ConfigT")
SubClient = TypeVar("SubClient")


class ApiClient:
    """
    The official Python client for interacting with the Crypticorn API.

    It is consisting of multiple microservices covering the whole stack of the Crypticorn project.
    """

    def __init__(
        self,
        api_key: str = None,
        jwt: str = None,
        base_url: BaseUrl = BaseUrl.PROD,
    ):
        self.base_url = base_url
        """The base URL the client will use to connect to the API."""
        self.api_key = api_key
        """The API key to use for authentication (recommended)."""
        self.jwt = jwt
        """The JWT to use for authentication (not recommended)."""

        self._service_classes: dict[Service, type[SubClient]] = {
            Service.HIVE: HiveClient,
            Service.TRADE: TradeClient,
            Service.KLINES: KlinesClient,
            Service.PAY: PayClient,
            Service.METRICS: MetricsClient,
            Service.AUTH: AuthClient,
        }

        self._services: dict[Service, SubClient] = {
            service: client_class(self._get_default_config(service))
            for service, client_class in self._service_classes.items()
        }

    @property
    def hive(self) -> HiveClient:
        """
        Entry point for the Hive AI API ([Docs](https://docs.crypticorn.com/api/?api=hive-ai-api)).
        """
        return self._services[Service.HIVE]

    @property
    def trade(self) -> TradeClient:
        """
        Entry point for the Trading API ([Docs](https://docs.crypticorn.com/api/?api=trading-api)).
        """
        return self._services[Service.TRADE]

    @property
    def klines(self) -> KlinesClient:
        """
        Entry point for the Klines API ([Docs](https://docs.crypticorn.com/api/?api=klines-api)).
        """
        return self._services[Service.KLINES]

    @property
    def metrics(self) -> MetricsClient:
        """
        Entry point for the Metrics API ([Docs](https://docs.crypticorn.com/api/?api=metrics-api)).
        """
        return self._services[Service.METRICS]

    @property
    def pay(self) -> PayClient:
        """
        Entry point for the Payment API ([Docs](https://docs.crypticorn.com/api/?api=payment-api)).
        """
        return self._services[Service.PAY]

    @property
    def auth(self) -> AuthClient:
        """
        Entry point for the Auth API ([Docs](https://docs.crypticorn.com/api/?api=auth-api)).
        """
        return self._services[Service.AUTH]

    async def close(self):
        """Close all client sessions."""
        for service in self._services.values():
            if hasattr(service.base_client, "close"):
                await service.base_client.close()

    def _get_default_config(
        self, service: Service, version: ApiVersion = ApiVersion.V1
    ):
        """
        Get the default configuration for a given service.
        """
        config_class = self._service_classes[service].config_class
        return config_class(
            host=f"{self.base_url}/{version}/{service}",
            access_token=self.jwt,
            api_key={aph.scheme_name: self.api_key} if self.api_key else None,
        )

    def configure(
        self,
        config: ConfigT,
        service: Service,
    ):
        """
        Update a sub-client's configuration by overriding with the values set in the new config.
        Useful for testing a specific service against a local server instead of the default proxy.

        :param config: The new configuration to use for the sub-client.
        :param service: The service to configure.

        Example:
        >>> async with ApiClient(base_url=BaseUrl.DEV, jwt=jwt) as client:
        >>>     client.configure(config=HiveConfig(host="http://localhost:8000"), client=client.hive)
        """
        assert Service.validate(service), f"Invalid service: {service}"
        client = self._services[service]
        new_config = client.config

        for attr in vars(config):
            new_value = getattr(config, attr)
            if new_value:
                setattr(new_config, attr, new_value)

        self._services[service] = type(client)(new_config)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()



"""
This module contains the admin router for the API.
It provides endpoints for monitoring the server and getting information about the environment.
ONLY ALLOW ACCESS TO THIS ROUTER WITH ADMIN SCOPES.
>>> app.include_router(admin_router, dependencies=[Security(auth_handler.combined_auth, scopes=[Scope.READ_ADMIN, Scope.WRITE_ADMIN])])
"""

import os
import pkg_resources
import threading
import time
import psutil
from fastapi import APIRouter, Query
from typing import Literal, Union
from crypticorn.common.logging import LogLevel
import logging

router = APIRouter(tags=["Admin"], prefix="/admin")

START_TIME = time.time()


@router.get("/log-level", status_code=200, operation_id="getLogLevel")
async def get_logging_level() -> LogLevel:
    """
    Get the log level of the server logger.
    """
    return LogLevel.get_name(logging.getLogger().level)


@router.get("/uptime", operation_id="getUptime", status_code=200)
def get_uptime(type: Literal["seconds", "human"] = "seconds") -> str:
    """Return the server uptime in seconds or human-readable form."""
    uptime_seconds = int(time.time() - START_TIME)
    if type == "seconds":
        return str(uptime_seconds)
    elif type == "human":
        return time.strftime("%H:%M:%S", time.gmtime(uptime_seconds))


@router.get("/memory", operation_id="getMemoryUsage", status_code=200)
def get_memory_usage() -> int:
    """
    Resident Set Size (RSS) in MB — the actual memory used by the process in RAM.
    Represents the physical memory footprint. Important for monitoring real usage.
    """
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    return round(mem_info.rss / (1024 * 1024), 2)


@router.get("/threads", operation_id="getThreads", status_code=200)
def get_threads() -> dict:
    """Return count and names of active threads."""
    threads = threading.enumerate()
    return {
        "count": len(threads),
        "threads": [t.name for t in threads],
    }


@router.get("/limits", operation_id="getContainerLimits", status_code=200)
def get_container_limits() -> dict:
    """Return container resource limits from cgroup."""
    limits = {}
    try:
        with open("/sys/fs/cgroup/memory/memory.limit_in_bytes") as f:
            limits["memory_limit_MB"] = int(f.read().strip()) / 1024 / 1024
    except Exception:
        limits["memory_limit_MB"] = "N/A"

    try:
        with open("/sys/fs/cgroup/cpu/cpu.cfs_quota_us") as f1, open(
            "/sys/fs/cgroup/cpu/cpu.cfs_period_us"
        ) as f2:
            quota = int(f1.read().strip())
            period = int(f2.read().strip())
            limits["cpu_limit_cores"] = quota / period if quota > 0 else "N/A"
    except Exception:
        limits["cpu_limit_cores"] = "N/A"

    return limits


@router.get("/dependencies", operation_id="getDependencies", status_code=200)
def list_installed_packages(
    include: list[str] = Query(
        default=None,
        description="List of dependencies to include in the response. If not provided, all installed packages will be returned.",
    )
) -> list:
    """Return a list of installed packages and versions."""
    return sorted(
        [
            {dist.project_name: dist.version}
            for dist in pkg_resources.working_set
            if include is None or dist.project_name in include
        ],
        key=lambda x: next(iter(x)),
    )



# coding: utf-8

"""
Crypticorn Auth API

OpenAPI compliant REST API built using tRPC with Express

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictStr, field_validator
from typing import Any, ClassVar, Dict, List, Optional
from typing import Optional, Set
from typing_extensions import Self


class GetApiKeys200ResponseInner(BaseModel):
    """
    GetApiKeys200ResponseInner
    """  # noqa: E501

    id: StrictStr = Field(description="ID of the API key")
    user_id: StrictStr = Field(description="User ID of the API key")
    scopes: List[StrictStr] = Field(description="Scopes of the API key")
    name: StrictStr = Field(description="Name of the API key")
    expires_at: Optional[datetime] = Field(
        default=None, description="Expiration time of the API key as a date"
    )
    created_at: datetime = Field(description="Creation time of the API key as a date")
    ip_whitelist: Optional[List[StrictStr]] = Field(
        default=None,
        description="IP addresses that can access the API key. If empty, the API key will be accessible from any IP address.",
    )
    __properties: ClassVar[List[str]] = [
        "id",
        "user_id",
        "scopes",
        "name",
        "expires_at",
        "created_at",
        "ip_whitelist",
    ]

    @field_validator("scopes")
    def scopes_validate_enum(cls, value):
        """Validates the enum"""
        for i in value:
            if i not in set(
                [
                    "read:predictions",
                    "read:hive:model",
                    "read:hive:data",
                    "write:hive:model",
                    "read:trade:bots",
                    "write:trade:bots",
                    "read:trade:exchangekeys",
                    "write:trade:exchangekeys",
                    "read:trade:orders",
                    "read:trade:actions",
                    "write:trade:actions",
                    "read:trade:exchanges",
                    "read:trade:futures",
                    "write:trade:futures",
                    "read:trade:notifications",
                    "write:trade:notifications",
                    "read:trade:strategies",
                    "write:trade:strategies",
                    "read:pay:payments",
                    "read:pay:products",
                    "write:pay:products",
                    "read:pay:now",
                    "write:pay:now",
                    "read:metrics:marketcap",
                    "read:metrics:indicators",
                    "read:metrics:exchanges",
                    "read:metrics:tokens",
                    "read:metrics:markets",
                    "read:sentiment",
                    "read:admin",
                    "write:admin",
                ]
            ):
                raise ValueError(
                    "each list item must be one of ('read:predictions', 'read:hive:model', 'read:hive:data', 'write:hive:model', 'read:trade:bots', 'write:trade:bots', 'read:trade:exchangekeys', 'write:trade:exchangekeys', 'read:trade:orders', 'read:trade:actions', 'write:trade:actions', 'read:trade:exchanges', 'read:trade:futures', 'write:trade:futures', 'read:trade:notifications', 'write:trade:notifications', 'read:trade:strategies', 'write:trade:strategies', 'read:pay:payments', 'read:pay:products', 'write:pay:products', 'read:pay:now', 'write:pay:now', 'read:metrics:marketcap', 'read:metrics:indicators', 'read:metrics:exchanges', 'read:metrics:tokens', 'read:metrics:markets', 'read:sentiment', 'read:admin', 'write:admin')"
                )
        return value

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of GetApiKeys200ResponseInner from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of GetApiKeys200ResponseInner from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "id": obj.get("id"),
                "user_id": obj.get("user_id"),
                "scopes": obj.get("scopes"),
                "name": obj.get("name"),
                "expires_at": obj.get("expires_at"),
                "created_at": obj.get("created_at"),
                "ip_whitelist": obj.get("ip_whitelist"),
            }
        )
        return _obj



# coding: utf-8

"""
Hive AI API

API for Hive AI model training and evaluation

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field
from typing import Any, ClassVar, Dict, List
from crypticorn.hive.client.models.coins import Coins
from crypticorn.hive.client.models.data_options import DataOptions
from crypticorn.hive.client.models.data_version_info import DataVersionInfo
from crypticorn.hive.client.models.feature_size import FeatureSize
from crypticorn.hive.client.models.target_info import TargetInfo
from typing import Optional, Set
from typing_extensions import Self


class DataInfo(BaseModel):
    """
    The complete data information for all versions, coins, feature sizes and targets
    """  # noqa: E501

    data: Dict[str, Dict[str, DataOptions]] = Field(
        description="The complete data information for all versions, coins, feature sizes and targets."
    )
    coins: List[Coins] = Field(
        description="The coins available on the latest data version."
    )
    feature_sizes: List[FeatureSize] = Field(
        description="The feature sizes available on the latest data version."
    )
    targets: List[TargetInfo] = Field(
        description="The targets available on the latest data version."
    )
    all_versions: List[DataVersionInfo] = Field(
        description="All ever existing data versions. Some may not be publicly available yet."
    )
    available_versions: List[DataVersionInfo] = Field(
        description="All publicly available data versions."
    )
    __properties: ClassVar[List[str]] = [
        "data",
        "coins",
        "feature_sizes",
        "targets",
        "all_versions",
        "available_versions",
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of DataInfo from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each value in data (dict)
        _field_dict = {}
        if self.data:
            for _key_data in self.data:
                if self.data[_key_data]:
                    _field_dict[_key_data] = self.data[_key_data].to_dict()
            _dict["data"] = _field_dict
        # override the default output from pydantic by calling `to_dict()` of each item in targets (list)
        _items = []
        if self.targets:
            for _item_targets in self.targets:
                if _item_targets:
                    _items.append(_item_targets.to_dict())
            _dict["targets"] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in all_versions (list)
        _items = []
        if self.all_versions:
            for _item_all_versions in self.all_versions:
                if _item_all_versions:
                    _items.append(_item_all_versions.to_dict())
            _dict["all_versions"] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in available_versions (list)
        _items = []
        if self.available_versions:
            for _item_available_versions in self.available_versions:
                if _item_available_versions:
                    _items.append(_item_available_versions.to_dict())
            _dict["available_versions"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of DataInfo from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "data": (
                    dict(
                        (
                            _k,
                            (
                                dict(
                                    (_ik, DataOptions.from_dict(_iv))
                                    for _ik, _iv in _v.items()
                                )
                                if _v is not None
                                else None
                            ),
                        )
                        for _k, _v in obj.get("data").items()
                    )
                    if obj.get("data") is not None
                    else None
                ),
                "coins": obj.get("coins"),
                "feature_sizes": obj.get("feature_sizes"),
                "targets": (
                    [TargetInfo.from_dict(_item) for _item in obj["targets"]]
                    if obj.get("targets") is not None
                    else None
                ),
                "all_versions": (
                    [DataVersionInfo.from_dict(_item) for _item in obj["all_versions"]]
                    if obj.get("all_versions") is not None
                    else None
                ),
                "available_versions": (
                    [
                        DataVersionInfo.from_dict(_item)
                        for _item in obj["available_versions"]
                    ]
                    if obj.get("available_versions") is not None
                    else None
                ),
            }
        )
        return _obj
